import logging
import os.path
import token, tokenize
from objects import *


class Token(object):
    """
    Holds Token information
    """
    def __init__(self, infos=None):
        if infos:
            self.type = infos[0]
            self.text = infos[1]
            self.s_line, self.s_col = infos[2]
            self.e_line, self.e_col = infos[3]
            self.line_text = infos[4]
            # print(self)
        else:
            self.type = None
            self.type = None
            self.text = None
            self.s_line = None
            self.s_col = None
            self.e_line = None
            self.e_col = None
            self.line_text = None

    def __str__(self):
        return "%10s %-14s %-20r %r" % (
            tokenize.tok_name.get(self.type, self.type),
            "%d.%d-%d.%d" % (self.s_line, self.s_col, self.e_line, self.e_col),
            self.text, self.line_text
        )

    def __repr__(self):
        return '<Token({})>'.format(str(self))


def _strip(text, quotes=False):
    """
    Strips indentation and quotes
    """
    text_ = ""

    for line in text.splitlines():
        line_ = line.strip()
        if quotes:
            line_ = line_.strip('"').strip("'")

        text_ += line_ + '\n'

    return text_


def _parse_class_parents(tokens):
    parents = list()
    parent = list()

    for infos in tokens:
        new_token = Token(infos)

        if new_token.type == token.NAME:
            parent.append(new_token.text)

        if new_token.type == token.OP and new_token.text in (',', ')'):
            parents.append('.'.join(parent))
            parent = list()

        if new_token.type == token.OP and new_token.text == ':':
            break

    return parents


def parse_arguments(tokens):
    arguments = list()
    argument = Argument()

    for infos in tokens: # TODO : _parse_value to deal with arg=[None, 1] for example
        new_token = Token(infos)

        if new_token.type == token.NAME and not argument.name:
            argument.name = new_token.text

        elif new_token.type in (token.NUMBER, token.STRING, token.NAME) and argument.name:
            argument.default = new_token.text

        elif new_token.type == token.OP and new_token.text in (',', ')'):
            if argument.name is not None:
                arguments.append(argument)
            argument = Argument()

        if new_token.type == token.OP and new_token.text == ':':
            break

    return arguments


def _parse_def(tokens):
    indent = None
    previous = Token()
    function_ = Function()


    for infos in tokens:
        new_token = Token(infos)

        if new_token.type == token.STRING and previous.type == token.INDENT:
            function_.docstring = Docstring(_strip(new_token.line_text, quotes=True))

        if new_token.type == token.INDENT and indent is None:
            indent = new_token.e_col

        if new_token.type == token.NAME and not function_.name:
            function_.name = new_token.text
            function_.arguments = parse_arguments(tokens)

        if new_token.type == token.DEDENT and new_token.e_col <= indent - 4:
            logging.info('Parsed function  : {}'.format(function_))
            return function_

        previous = new_token


def _parse_class(tokens):
    logging.info('New class')
    indent = None
    new_token = Token(tokens.next())
    previous = new_token

    new_class = Class()
    new_class.name = new_token.text
    new_class.parents = _parse_class_parents(tokens)

    for infos in tokens:
        new_token = Token(infos)

        if new_token.type == token.INDENT and indent is None:
            indent = new_token.e_col

        if new_token.type == token.STRING and previous.type == token.INDENT:
            new_class.docstring = Docstring(_strip(new_token.line_text, quotes=True))

        if new_token.type == token.NAME and new_token.text == 'def':
            function_ = _parse_def(tokens)
            if function_.name == '__init__':
                new_class.constructor = function_
            else:
                new_class.content.append(function_)

        if new_token.type == token.DEDENT and new_token.e_col <= indent - 4:
            logging.info('Parsed class     : {}'.format(new_class))
            return new_class

        previous = new_token


def _parse_module(tokens, module):
    previous_type = token.INDENT

    for infos in tokens:
        new_token = Token(infos)

        if previous_type == token.INDENT and new_token.type == token.STRING:
            module.docstring = Docstring(_strip(new_token.line_text, quotes=True))

        if new_token.type == token.NAME and new_token.text == 'class':
            new_class = _parse_class(tokens)
            module.content.append(new_class)

        if new_token.type == token.NAME and new_token.text == 'def':
            new_function = _parse_def(tokens)
            module.content.append(new_function)

        previous_type = new_token.type

    logging.info('Parsed module    : {}'.format(module))
    return module


def parse_module(filepath):
    module = Module()
    module.name = os.path.basename(filepath)

    with open(filepath, 'r') as f_source:
        tokens = tokenize.generate_tokens(f_source.readline)
        module = _parse_module(tokens, module)

    return module



"""
Objects generated by the parser
"""

class _Object(object):
    """
    Root class of objects

    Has a `.name`
    """
    def __init__(self):
        self.name = None

    def __repr__(self):
        return "<{}({})>".format(self.__class__.__name__, str(self))


class Docstring(_Object):
    """Holds a docstring

    Has a `.content`"""
    def __init__(self, content):
        _Object.__init__(self)
        self.content = content

    def __str__(self):
        return self.content


class Module(_Object):
    """Module

    Has a `.docstring` (defaults to `None`) and a `.content` (list of _Object subclass instances)"""
    def __init__(self):
        _Object.__init__(self)
        self.docstring = None
        self.content = list()

    def __str__(self):
        return self.name


class Class(_Object):
    """Class

    Has `.parents` (name of parent classes), `.docstring`, `.constructor` and `.content`
    """
    def __init__(self):
        _Object.__init__(self)
        self.parents = list()
        self.docstring = None
        self.constructor = None
        self.content = list()

    def __str__(self):
        if self.constructor is not None:
            return '{}({})'.format(self.name, ', '.join(str(a) for a in self.constructor.arguments))
        return '{}()'.format(self.name)


class Argument(_Object):
    """Argument to functions

    Has a `.default` value (that can be `None`)
    """
    def __init__(self):
        _Object.__init__(self)
        self.default = ""

    def __str__(self):
        if self.default:
            return '{}={}'.format(self.name, self.default)
        return self.name


class Function(_Object):
    """Function of class method

    Has a `.name`, `.docstring` and a list of `.arguments`"""
    def __init__(self):
        _Object.__init__(self)
        self.docstring = None
        self.arguments = list()

    def __str__(self):
        if self.arguments:
            return '{}({})'.format(self.name, ', '.join([str(a) for a in self.arguments]))
        else:
            return self.name + '()'



import logging
from frangidoc.objects import *


_MODULE = """{title} Module {name}

{docstring}

{content}
"""

_CLASS = """{title} class {name}

```python
{signature}
```

{docstring}

{content}
"""

_METHOD = """{title} {class_name}.{name}

```python
{signature}
```

{docstring}
"""

_FUNCTION = """{title} {name}

```python
{signature}
```

{docstring}
"""


def _render_docstring(docstring, parent, level):
    logging.info('Rendering docstring : {}'.format(parent.name))
    lines = docstring.content.splitlines()
    table = ['| Argument | Role |', '|---|---|']
    output = list()
    return_info = list()

    for line in lines:
        if line.startswith(':param '):
            name = line[6:].split(':')[0]
            info = ':'.join(line[6:].split(':')[1:])

            table.append('| `{}` | {} |'.format(name, info))

        elif line.startswith(':return:'):
            info = line[8:]
            if info:
                return_info.append('')
                return_info.append('**Returns** : {}'.format(info))

        else:
            output.append(line)

    if len(table) > 2:
        return '\n'.join(output + table + return_info)

    return '\n'.join(output + return_info)


def _title(level):
    """Markdown title"""
    return '#' * (level + 1) + ' '


def _render_module(module, parent, level):
    """Renders a module to Markdown"""
    logging.info('Rendering module    : {}'.format(module.name))
    title = _title(level)
    name = module.name
    docstring = render(module.docstring, module) if module.docstring else ''
    content = '\n'.join(render(item, None, level + 2) for item in module.content)

    logging.info('Rendering done      : {}'.format(module.name))

    return _MODULE.format(
        title=title,
        name=name,
        docstring=docstring,
        content=content
    ).strip()


def _render_class(class_, parent, level):
    """Renders a class to Markdown"""
    logging.info('Rendering class     : {}'.format(class_.name))
    title = _title(level)
    name = class_.name
    signature = class_
    docstring = class_.docstring if class_.docstring else ''
    content = '\n'.join(render(item, class_, level + 1) for item in class_.content)

    logging.info('Rendering done      : {}'.format(class_.name))

    return _CLASS.format(
        title=title,
        name=name,
        signature=signature,
        docstring=docstring,
        content=content
    ).strip()


def _render_function(function, parent, level):
    logging.info('Rendering function  : {}'.format(function.name))
    title = _title(level)
    docstring = render(function.docstring, function) if function.docstring else ''

    if isinstance(parent, Class):
        class_name = parent.name
        name =  function.name
        signature = parent.name + '.' + str(function)

        return _METHOD.format(
            title=title,
            class_name=class_name,
            name=name,
            signature=signature,
            docstring=docstring
        )

    name = function.name
    signature = str(function)

    return _FUNCTION.format(
        title=title,
        name=name,
        signature=signature,
        docstring=docstring
    )


def render(item, parent=None, level=0):
    if isinstance(item, Module):
        return '\n' + _render_module(item, parent, level)

    if isinstance(item, Class):
        return '\n' + _render_class(item, parent, level)

    if isinstance(item, Function):
        return '\n' + _render_function(item, parent, level)

    if isinstance(item, Docstring):
        return '\n' + _render_docstring(item, parent, level)



import os
import sys
import copy
import stat
import errno
import shutil
import logging
import tempfile
import subprocess
import git
import yaml
from frangidoc.parser import parse_module
from frangidoc.renderer import render


def _handle_remove_read_only(func, path, exc):
  excvalue = exc[1]
  if func in (os.rmdir, os.remove) and excvalue.errno == errno.EACCES:
      os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO) # 0777
      func(path)
  else:
      raise


def _save(filepath, content):
    if not os.path.isdir(os.path.dirname(filepath)):
        os.makedirs(os.path.dirname(filepath))
    with open(filepath, "w+") as f_file:
        f_file.write(content)


def generate_and_save(module_filepath, output_filepath=None):
    logging.info('Processing file : {}'.format(module_filepath))

    module = parse_module(module_filepath)
    page = render(module)

    logging.info('Processing done !')

    _save(output_filepath, page)

    logging.info("Output file written : %s" % output_filepath)


def _make_pythonpath_envvar():
    if os.name == 'nt':
        return ';'.join(sys.path)
    return ':'.join(sys.path)


def argparse_and_save(package_name, module_filepath, output_filepath):
    logging.info(" -> Generating Markdown from CLI {}".format(package_name))

    command = "python {} -h".format(module_filepath)
    env = os.environ.copy()
    env['PYTHONPATH'] = _make_pythonpath_envvar()

    try:
        content = subprocess.check_output(command, shell=True, env=env)
    except Exception as e:
        logging.warning("    Error while executing '{}' :".format(command))
        logging.warning(e)
        return

    lines = [
        '# {} : CLI'.format(package_name),
        '',
        'Here is the output generated by running `python -m {} -h` :'.format(package_name),
        '',
        '```text'
    ]

    lines += content.splitlines()
    lines.append('```')

    content = '\n'.join(lines)
    _save(output_filepath, content)

    logging.info("Output file written : %s" % output_filepath)


def clone_and_generate(repository_url, output_directory, cleanup=True):
    # TODO : range moi ca !
    repo_name = os.path.basename(repository_url).replace('.git', '')
    temp_folder = tempfile.mkdtemp(prefix="frangidoc-{}.".format(repo_name))
    config_filepath = os.path.join(temp_folder, '.frangidoc.yml')

    try:
        git.Repo.clone_from(repository_url, temp_folder)
    except git.GitCommandError as e:
        logging.warning("Impossible to clone {repo_url}".format(repo_url=repository_url))
        logging.warning(e)
        return False

    logging.info("Cloned {repo_url} to {temp_folder}".format(repo_url=repository_url, temp_folder=temp_folder))

    if not os.path.isfile(config_filepath):
        logging.warning("Could not find .frangidoc.yml, aborting")
        return False

    with open(config_filepath, 'r') as f_config:
        config = yaml.load(f_config)

    logging.info("Loaded .frangidoc.yml")

    output_folder = os.path.join(output_directory, config['title'])

    if os.path.isdir(output_folder):
        shutil.rmtree(output_folder, ignore_errors=False, onerror=_handle_remove_read_only)

    os.makedirs(output_folder)

    sys_path_backup = copy.deepcopy(sys.path)
    environment_backup = copy.deepcopy(os.environ)

    environment = config.get('environment', dict())
    for key, value in environment.items():
        value = value.replace(':', ';')
        logging.info("Adding to {} : {}".format(key, value))

        os.environ[key] = value

        if key.upper() == 'PYTHONPATH':
            for path in value.split(';'):
                sys.path.append(path)

    sys_path_inter_backup = copy.deepcopy(sys.path)

    for module_filepath in config.get('modules', list()):
        module_fullpath = os.path.join(temp_folder, module_filepath)
        relative_path, filename = os.path.split(module_filepath)
        path = os.path.join(temp_folder, relative_path)
        package_name = os.path.basename(relative_path)
        leaf = '/'.join([step for step in os.path.split(relative_path) if step != ''][1:])
        name = filename.replace('.py', '')

        # todo Deplacer cette logique dans 'generate_and_save()'
        # todo (pour que l'api soit consistante entre 'git' et 'module')
        if name == '__init__':
            name = package_name

        if name in ('__init__', '__main__'):
            path = os.path.dirname(path)

        if name == '__main__':
            output_filename = os.path.join(output_folder, leaf, package_name + ' (CLI).md')
            argparse_and_save(package_name, module_fullpath, output_filename)

        else:
            output_filename = os.path.join(output_folder, leaf, name + '.md')
            generate_and_save(module_fullpath, output_filename)

        sys.path = copy.deepcopy(sys_path_inter_backup)

    sys.path = copy.deepcopy(sys_path_backup)
    os.environ = environment_backup

    for markdown_filepath in config.get('markdown', list()):
        fullpath = os.path.join(temp_folder, markdown_filepath)

        if not os.path.isfile(fullpath):
            logging.info("File not found {}".format(markdown_filepath))

        output = os.path.join(output_folder, markdown_filepath)
        shutil.copy2(fullpath, output)
        logging.info("Copied {}".format(markdown_filepath))

    if cleanup:
        logging.info("Cleaning up {}".format(temp_folder))
        shutil.rmtree(temp_folder, ignore_errors=False, onerror=_handle_remove_read_only)

    return True
